#include <iostream>
#include <filesystem>
#include <fcntl.h>
#include <unistd.h>

#include <pwd.h>
#include <grp.h>
#include <cstring>

#include <thread>

#include "BSize.h"

const unsigned char* const g_buffer{
        static_cast<unsigned char*>(aligned_alloc(512, 16777216))};

int performTest(const std::filesystem::path& path_to_write,
        const unsigned char* const buffer,
        const ssize_t buffer_size,
        const std::chrono::milliseconds test_duration)
{
    // Create temporary file if provided path is directory
    const bool create_tmp = std::filesystem::is_directory(path_to_write);
    ssize_t written = 0;

    auto file_descriptor = open(path_to_write.c_str(),
            O_WRONLY | O_DIRECT | (create_tmp ? O_TMPFILE : O_CREAT));
    if (file_descriptor<0) {
        std::cerr << "Cannot create temporary file " << std::strerror(errno) << std::endl;
        return -2;
    }

    auto start_time = std::chrono::steady_clock::now();
    while ((std::chrono::steady_clock::now()-start_time)<test_duration) {
        auto written_l = write(file_descriptor, buffer, buffer_size);
        if (written_l<=0) {
            std::cerr << "Cannot write temporary file " << std::strerror(errno) << std::endl;
            return -2;
        }
        written += written_l;
    }
    // Wait for all data to be written into the file
    if (fdatasync(file_descriptor)) {
        std::cerr << "Cannot sync/save temporary file " << std::strerror(errno) << std::endl;
        return -2;
    }

    auto end_time = std::chrono::steady_clock::now();
    auto duration{end_time-start_time};
    typeof(duration) converted_writing_duration{std::chrono::seconds{1}};
    long long speed = (written*converted_writing_duration.count())/duration.count();

    std::cout << format(written) << format(speed)
              << "\b/s Buffer:" << format(buffer_size, false)
              << "Time:" << std::chrono::duration_cast<std::chrono::milliseconds>(end_time-start_time).count() << "ms"
              << std::endl;

    close(file_descriptor);
    if (not create_tmp) {
        remove(path_to_write.c_str());
    }

    return 0;
}

int performTest(const std::filesystem::path& path_to_write,
        const unsigned char* const buffer,
        const ssize_t buffer_size,
        const ssize_t file_size,
        const ssize_t total_io_operations_size)
{
    // Create temporary file if provided path is directory
    const bool create_tmp = std::filesystem::is_directory(path_to_write);
    ssize_t total_written = 0;
    ssize_t file_written = 0;

    auto start_time = std::chrono::steady_clock::now();
    while (total_written<total_io_operations_size) {
        auto file_descriptor = open(path_to_write.c_str(),
                O_WRONLY | O_DIRECT | (create_tmp ? O_TMPFILE : O_CREAT));
        if (file_descriptor<0) {
            std::cerr << "Cannot create temporary file " << std::strerror(errno) << std::endl;
            return -2;
        }

        while (file_written<file_size) {
            auto written_l = write(file_descriptor, buffer, buffer_size<file_size ? buffer_size : file_size);
            if (written_l<=0) {
                std::cerr << "Cannot write temporary file " << std::strerror(errno) << std::endl;
                return -2;
            }
            file_written += written_l;
        }

        total_written += file_written;
        file_written = 0;

        // Wait for all data to be written into the file
        if (fdatasync(file_descriptor)) {
            std::cerr << "Cannot sync/save temporary file " << std::strerror(errno) << std::endl;
            return -2;
        }

        close(file_descriptor);
        if (not create_tmp) {
            remove(path_to_write.c_str());
        }
    }

    auto end_time = std::chrono::steady_clock::now();
    auto duration{end_time-start_time};
    typeof(duration) converted_writing_duration{std::chrono::seconds{1}};
    long long speed = (total_written*converted_writing_duration.count())/duration.count();

    std::cout << format(total_written) << format(speed)
              << "\b/s B/File:" << format(file_size)
              << "Buffer:" << format(buffer_size, false)
              << "Time:" << std::chrono::duration_cast<std::chrono::milliseconds>(end_time-start_time).count() << "ms"
              << std::endl;

    return 0;
}

int main(int argc, char* argv[])
{
    std::filesystem::path path_to_write{std::filesystem::current_path()};
    std::cout << "Usage: swiftMark /path/to/mounted/disc [user] [group]" << std::endl;
    if (argc>=1 && argv[1]!=nullptr) {
        path_to_write = argv[1];
    }

    std::cout << "Testing speed on: " << path_to_write << std::endl;

    if (argc>=2 && argv[2]!=nullptr) {
        auto pw = getpwnam(argv[2]);
        if (pw==nullptr) {
            std::cerr << "Cannot find user " << argv[2] << std::endl;
            return -1;
        }
        if (setgid(pw->pw_gid)) {
            auto gpw = getgrnam(argv[2]);
            std::cerr << "Cannot change group to " << gpw->gr_name << std::endl;
            return -1;
        }
        if (setuid(pw->pw_uid)) {
            std::cerr << "Cannot change user to " << argv[2] << std::endl;
            return -1;
        }
    }

    if (argc>=3 && argv[3]!=nullptr) {
        auto pw = getgrnam(argv[3]);
        if (pw==nullptr) {
            std::cerr << "Cannot find group " << argv[3] << std::endl;
            return -1;
        }
        if (setgid(pw->gr_gid)) {
            std::cerr << "Cannot change group to " << argv[3] << std::endl;
            return -1;
        }
    }

    auto user_pw = getpwuid(geteuid());
    if (user_pw==nullptr) {
        std::cerr << "Cannot get user name from " << geteuid() << std::endl;
        return -1;
    }

    auto group_pw = getgrgid(getgid());
    if (group_pw==nullptr) {
        std::cerr << "Cannot get group name from " << getgid() << std::endl;
        return -1;
    }

    std::cout << "Running as user: " << user_pw->pw_name << " group " << group_pw->gr_name << std::endl;;

    std::cout << std::endl << "Running buffer tests" << std::endl;
    // Fill SSD SLC sector for actual (without SLC buffer) TLC write speed.
    performTest(path_to_write, g_buffer, 16_MiB, std::chrono::seconds{2});
    for (int i = 1; i<16; i += 2) {
        // If buffer have known size, one can write to a disc then perform test.
        // Longer test duration can provide better non buffered data writing speed.
        performTest(path_to_write, g_buffer, 512 << i, std::chrono::seconds{1});
        // Sleep so disc buffer can be dumped (SSD SLC).
        //std::this_thread::sleep_for(std::chrono::seconds (5));
    }

    std::cout << std::endl << "Running file tests" << std::endl;

    performTest(path_to_write, g_buffer, 16_MiB, 1_kiB, 2_MB+500_kB);
    performTest(path_to_write, g_buffer, 16_MiB, 16_kiB, 50_MB);
    performTest(path_to_write, g_buffer, 16_MiB, 1_MiB, 1_GB);
    performTest(path_to_write, g_buffer, 16_MiB, 100_MB, 2_GB);

    return 0;
}
